# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import copy
import inspect
import logging
import re

from collections import OrderedDict
from typing import Callable, Dict, List, AnyStr, Tuple

import forge
import inflection

from factory.base import Factory
from graphql import (
    GraphQLNonNull,
    GraphQLInt,
    GraphQLFloat,
    GraphQLString,
    GraphQLBoolean,
    GraphQLField,
    GraphQLArgument,
    GraphQLScalarType,
    GraphQLInputObjectField,
    GraphQLList,
    GraphQLInputObjectType,
    GraphQLObjectType,
)
from gql import Client, gql
from gql_query_builder import GqlQuery

from qctrlapiclient.api import Api
from qctrlapiclient.model import Model
from qctrlcommons.node.registry import NODE_REGISTRY
from qctrl.result import Result
from qctrl.static.services import (
    CoherentSimulationService,
    ColoredNoiseSimulationService,
    FilterFunctionService,
    NoiseCharacterizationService,
    NoiseReconstructionService,
    NoiseFreeMetricsService,
    OptimalOptimizedSystemService,
    PiecewisePowerLawNoiseSpectralDensityService,
    QuasiStaticFunctionService,
    RobustOptimizedSystemService,
    T1DecayAssessmentService,
    FlexibleOptimizedSystemService,
)
from .config import get_config
from .doc import FunctionDocstring, FunctionArgument, ReturnType
from .namespaces import (
    BaseNamespace,
    create_factory_namespace,
    create_function_namespace,
    create_model_namespace,
    create_operation_namespace,
    create_service_namespace,
    create_type_namespace,
)
from .query_builder import QueryBuilder
from .signature_builder import SignatureBuilder
from .type_registry import TypeRegistry
from .model_factory import ModelFactory
from .factory_builder import FactoryBuilder
from .model_builder import ModelBuilder

LOGGER = logging.getLogger(__name__)


SERVICE_REGISTRY = [
    CoherentSimulationService,
    ColoredNoiseSimulationService,
    FilterFunctionService,
    NoiseCharacterizationService,
    NoiseReconstructionService,
    NoiseFreeMetricsService,
    OptimalOptimizedSystemService,
    PiecewisePowerLawNoiseSpectralDensityService,
    QuasiStaticFunctionService,
    RobustOptimizedSystemService,
    T1DecayAssessmentService,
    FlexibleOptimizedSystemService,
]


def is_valid_function_mutation(mutation_field: GraphQLField) -> bool:
    """
    Checks if the mutation field corresponds to a valid
    function which should be available to the user. Mutations
    that should be exposed have a single argument named
    `input`.
    """
    args = mutation_field.args
    return len(args) == 1 and "input" in args


def create_function(
    client: Client,
    mutation_name: str,
    mutation_field: GraphQLField,
    type_registry: TypeRegistry = None,
) -> Callable:
    """
    Returns a function which can be used to execute
    a GraphQL mutation.
    """
    query_builder = QueryBuilder(mutation_name, mutation_field)

    docstring = FunctionDocstring(
        description=mutation_field.description,
        returns=ReturnType("qctrl.result.Result"),
    )

    def func(*args, **kwargs) -> Result:  # pylint:disable=unused-argument
        query = query_builder(args, kwargs)
        LOGGER.debug("query:%s", query)
        response = client.execute(query)
        LOGGER.debug("response:%s", response)

        if mutation_name not in response:
            LOGGER.error("response:%s", response)
            raise ValueError(f"response missing mutation name: {mutation_name}")

        result = Result(response[mutation_name])
        LOGGER.info(
            "Remote function %s called - Action ID: %s", result.name, result.action_id
        )
        result.wait_for_completion()
        return result

    input_fields = query_builder.get_input_fields()

    # add docstring
    for field_name, field in input_fields.items():
        docstring.args.append(
            FunctionArgument(field_name, description=field.description)
        )

    func.__doc__ = str(docstring)

    # add signature
    signature = SignatureBuilder(input_fields, type_registry)
    func.__name__ = inflection.underscore(mutation_name)
    return signature(func)


def _get_factory_alias(factory: Factory) -> str:
    """
    Returns the attribute name that the
    factory should be accessible at.
    """
    attr = re.sub("Factory$", "", factory.__name__)
    attr = inflection.underscore(attr)
    return inflection.pluralize(attr)


def _get_api_models(api: Api) -> List[Tuple[str, Model]]:
    """
    Returns the API models.
    """
    models = []

    for attr, value in inspect.getmembers(api):
        if isinstance(value, Model):
            models.append((attr, value))

    return models


# pylint:disable=too-many-locals
def build_namespaces(api: Api, gql_client: Client):
    """
    Builds features and organizes them into namespaces.

    Args:
        api (Api): the JSON:API Api object
        gql_client (Client): the GraphQL client object

    Returns:
        tuple: the BaseNamespace objects in the following
        order - factory, function, model, operation, service
        type.

    """
    factory_namespace = create_factory_namespace()
    function_namespace = create_function_namespace()
    model_namespace = create_model_namespace()
    operation_namespace = create_operation_namespace()
    service_namespace = create_service_namespace()
    type_namespace = create_type_namespace()

    # models and factories
    factory_builder = FactoryBuilder(api)
    model_builder = ModelBuilder(api)

    ignored_factories = get_config("IGNORE_FACTORIES", [])

    for resource_name, api_model_instance in _get_api_models(api):

        # for every model, create a class of that mode e.g. class System
        model_class = model_builder(resource_name, api_model_instance)
        LOGGER.debug("model_class: %s", model_class)

        # create an instance of the class which can be used to create models
        model_instance = model_class(
            resource_name,
            api_model_instance._session,  # pylint:disable=protected-access
        )

        # register model
        model_namespace.extend({resource_name: model_instance})
        LOGGER.info("registered model: %s", resource_name)

        # dont create factories for these models
        if resource_name in ignored_factories:
            LOGGER.debug("not creating factory for resource: %s", resource_name)
            continue

        # build factories for model
        factories = factory_builder(resource_name, model_instance)
        LOGGER.debug("factories: %s", factories)

        for factory in factories:
            attr = _get_factory_alias(factory)
            factory_namespace.extend({attr: factory})
            LOGGER.info("registered factory: %s, %s", attr, factory)

    # services
    for service_cls in SERVICE_REGISTRY:
        service_namespace.extend(
            {service_cls._label: service_cls(api)}  # pylint:disable=protected-access
        )

    # operations
    for node_cls in NODE_REGISTRY.as_list():
        func = node_cls.create_pf()
        operation_namespace.extend_functions(func)

    # functions and types
    type_registry = TypeRegistry()
    mutation_type = gql_client.schema.get_mutation_type()

    for name, field in mutation_type.fields.items():
        if is_valid_function_mutation(field):
            func = create_function(gql_client, name, field, type_registry)
            function_namespace.extend_functions(func)

    type_namespace.extend(type_registry.get_type_map())

    return (
        factory_namespace,
        function_namespace,
        model_namespace,
        operation_namespace,
        service_namespace,
        type_namespace,
    )


__all__ = [
    "build_namespaces",
]
