# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring

from dataclasses import dataclass, field as dataclass_field
from datetime import datetime
from functools import wraps
from typing import Callable, List, Optional, Union

import forge
import numpy
from qctrlapiclient.resourceobject import ResourceObject
from qctrlcommons.utils import get_nested_value

from qctrl.utils import get_resource_schema, normalize_encoded_field_name


def _period_join(period_list: list):
    """Utility function to join a list by ". " ensuring that
    there are no repeated periods and that the
    final item ends in a period.
    """
    period_list = [x.strip(".") for x in period_list]
    return ". ".join(period_list) + "."


class FieldType:  # pylint:disable=too-few-public-methods
    """Describes the type of a schema field.
    """

    def __init__(self, description: str, type_hint=None, constraints: dict = None):
        self.constraints = constraints
        self.description = description
        self.type_hint = type_hint

    def is_match(self, schema: dict) -> bool:
        """Checks if the field schema corresponds to this field type using the constraints.
        """
        if not self.constraints:
            return False

        for key, expected_value in self.constraints.items():
            if expected_value != schema.get(key):
                return False

        return True


class DocField:  # pylint:disable=too-few-public-methods
    """Handles standard vs encoded doc fields.
    """

    def __new__(cls, name, *args, **kwargs):
        if name.endswith("_encoded"):
            return EncodedDocField(name, *args, **kwargs)

        return StandardDocField(name, *args, **kwargs)


class StandardDocField:
    """Represents documentation for a standard field.
    """

    # doc types
    DATETIME = "datetime"
    STRING = "str"
    INTEGER = "int"
    FLOAT = "float"
    DICT = "dict"
    LIST = "list"
    BOOLEAN = "bool"

    # field type definitions
    _types = [
        FieldType(DATETIME, datetime, {"type": "string", "format": "date-time"},),
        FieldType(STRING, str, {"type": "string"}),
        FieldType(INTEGER, int, {"type": "integer"}),
        FieldType(FLOAT, float, {"type": "number"}),
        FieldType(DICT, dict, {"type": "object"}),
        FieldType(LIST, list, {"type": "array"}),
        FieldType(BOOLEAN, bool, {"type": "boolean"}),
    ]

    # formatting of any validation included
    # in the schema; the schema keys and self
    # will be available for message formatting
    _validation_info = {
        # (schema keys, message-or-function)
        ("minimum", "maximum"): "Between {minimum} and {maximum}",
        ("minimum",): "Greater than or equal to {minimum}",
        ("maximum",): "Less than or equal to {maximum}",
        (
            "enum",
        ): lambda self, enum: self._format_enum(  # pylint:disable=protected-access
            enum
        ),
    }

    def __init__(self, name, schema, required=False):
        self.name = name
        self.schema = schema
        self.required = required
        self.description = self._get_description()
        self._field_type = self._get_field_type()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return (
            self.name == other.name
            and self.schema == other.schema
            and self.required == other.required
        )

    def clone(self, **overrides):
        """Clone the instance.
        """
        kwargs = dict(name=self.name, schema=self.schema, required=self.required)

        kwargs.update(overrides)
        return self.__class__(**kwargs)

    @property
    def type_info(self):
        """class property to get description from field_type
        """
        if self._field_type:
            return self._field_type.description

        return "Unknown type"

    @property
    def type_hint(self):
        """class property to get type_hint from field_type
        """
        if self._field_type:
            return self._field_type.type_hint

        return None

    def is_read_only(self) -> bool:
        """Determines if the attribute is read only
        """
        return self.schema.get("readOnly", False)

    def _get_description(self) -> str:
        """Fetch the description and append validation rules.
        """
        parts = []

        description = self.schema.get("description", None)
        if description:
            parts = [description]

        if self.required:
            parts.append("Required")

        validation = self._get_validation()
        if validation:
            parts.append(validation)

        if parts:
            return _period_join(parts)

        return ""

    def _get_validation(self):
        """Returns documentation for any supported validation constraints.
        """
        for required_keys, message in self._validation_info.items():
            values = {}

            # check for required keys
            # in the schema
            for key in required_keys:
                if key in self.schema:
                    values[key] = self.schema[key]
                else:
                    values = False
                    break

            # if any not found, go to next combination
            if not values:
                continue

            # return the formatted error message
            if callable(message):
                return message(self, **values)

            return message.format(self=self, **values)

    @staticmethod
    def _format_enum(enum: list) -> str:
        return "One of: " + ", ".join(map(str, enum))

    def _get_field_type(self) -> FieldType:
        """Returns the corresponding FieldType. Returns None if type cannot be determined.
        """
        for field_type in self._types:
            if field_type.is_match(self.schema):
                return field_type

        # if no match found, check if enum field and get
        # implied type from enum option
        enum = self.schema.get("enum", [])

        if enum:
            first_item = enum[0]

            for field_type in self._types:
                if isinstance(first_item, field_type.type_hint):
                    return field_type

        return None

    def __str__(self):
        parts = [self.name]

        if self.type_info:
            parts.append(f" ({self.type_info})")

        if self.description:
            parts.append(f": {self.description}")

        return "".join(parts).strip()


class EncodedDocField(StandardDocField):
    """Represents documentation for an encoded field.
    """

    NUMPY_ARRAY = "numpy.ndarray"

    # special encoded field overrides
    _encoded_types = {
        # dict -> encoded numpy array
        StandardDocField.DICT: FieldType(NUMPY_ARRAY, numpy.ndarray),
        # list -> encoded numpy array or list
        StandardDocField.LIST: FieldType(
            f"{StandardDocField.LIST} or {NUMPY_ARRAY}", Union[list, numpy.ndarray]
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # modify field name
        self.name = normalize_encoded_field_name(self.name)

    def _get_field_type(self):
        """Returns the modified schema type for encoded values.
        """
        field_type = super()._get_field_type()

        if field_type and field_type.description in self._encoded_types:
            return self._encoded_types[field_type.description]

        return field_type


class RelationshipDocField(StandardDocField):
    """Represents documentation for a related resource.
    """

    _relationship_types = {
        "object": FieldType(
            "qctrlapiclient.resourceobject.ResourceObject", ResourceObject
        ),
        "array": FieldType(
            "List of qctrlapiclient.resourceobject.ResourceObject", list
        ),
    }

    def is_read_only(self) -> bool:
        if super().is_read_only():
            return True

        if self.type_hint == list:
            return True

        return False

    def _get_field_type(self):
        _type = get_nested_value(self.schema, "properties.data.type")
        return self._relationship_types.get(_type)

    def _get_description(self):
        parts = []
        if self.required:
            parts.append("Required")

        if parts:
            return _period_join(parts)

        return ""


class DocFieldCache:  # pylint:disable=too-few-public-methods
    """Cache for storing DocFields for each resource.
    """

    def __init__(self, schema: dict):
        self._schema = schema
        self._data = {}

    def get_fields(self, resource_name: str) -> list:
        """A function to get a lists of field based on the resource name.
        """
        if resource_name not in self._data:
            self._create_fields(resource_name)

        return self._data[resource_name]

    def _create_fields(self, resource_name: str):
        schema = get_resource_schema(self._schema, resource_name)

        fields = []

        # add regular attributes
        required_fields = get_nested_value(schema, "properties.attributes.required", [])
        properties = get_nested_value(schema, "properties.attributes.properties", {})

        for name, field_schema in properties.items():
            fields.append(DocField(name, field_schema, name in required_fields))

        # add relationships
        relationships = get_nested_value(
            schema, "properties.relationships.properties", {}
        )

        for name, relationship_schema in relationships.items():
            fields.append(RelationshipDocField(name, relationship_schema))

        # sort all fields
        self._data[resource_name] = sorted(fields, key=lambda f: f.name)


class DocGenerator:  # pylint:disable=too-few-public-methods
    """Generates documentation for the schema.
    """

    def __init__(self, schema: dict):
        """Provide the Open API schema.
        """
        self._schema = schema
        self._cache = DocFieldCache(schema)

    def get_fields(
        self, resource_name: str, only=None, write_only=False, overrides=None
    ) -> list:
        """Returns the list of DocField objects for fields in the schema. Optionally ignore type
        for typed factories.
        """
        fields = []

        for field in self._cache.get_fields(resource_name):

            if only and field.name not in only:
                continue

            if write_only and field.is_read_only():
                continue

            if overrides and field.name in overrides:
                field = field.clone(**overrides[field.name])

            fields.append(field)

        return fields

    @staticmethod
    def get_docstring(
        fields: list,
        section_title: str,
        extra_text: str = None,
        return_object: str = None,
    ) -> str:
        """A function to get a list of doc string from fields, section title,
        (optional: extra text and return object).
        """
        lines = []

        # add extra text (if any)
        if extra_text:
            lines.append(extra_text.strip())
        else:
            lines.append("")

        # add args section
        lines.append(f"{section_title}:")

        for field in fields:
            lines.append(f"  {field}")

        # add returns section
        if return_object:
            lines.append("Returns:")
            lines.append(f"  {return_object}: A {return_object} object.")

        # add raises section
        lines.append("Raises:")
        lines.append("  QctrlApiException")

        return "\n\n".join(lines)

    @staticmethod
    def get_signature_wrapper(fields: list) -> Callable:
        """A function to use forge python package to get the function signatures.
        """
        args = [forge.arg("cls")]

        for field in fields:
            args.append(
                forge.arg(
                    field.name,
                    type=field.type_hint
                    or forge._marker.empty,  # pylint:disable=protected-access
                    default=None,
                )
            )

        return forge.sign(*args)


class Indent:  # pylint:disable=too-few-public-methods
    """
    Handles indentation of docstrings.

    Attributes:
        count (int): how many indentations should be made.
        extra_for_new_lines (int): how many extra indentations
            should apply to new lines within the given string.
        spacing: the indentation string.
    """

    def __init__(
        self, count: int = 1, extra_for_new_lines: int = 0, spacing: str = "    "
    ):
        self.count = count
        self.extra_for_new_lines = extra_for_new_lines
        self.spacing = spacing

    def _apply_spacing(self, line: str, count: int) -> str:
        """
        Indents the line by the given count.
        """
        for _ in range(count):
            line = self.spacing + line

        return line

    def __call__(self, string: str) -> str:
        """
        Indents a single or multi line string.
        """
        lines = string.strip().split("\n")

        first_line = True

        for index, line in enumerate(lines):
            if first_line:
                line = self._apply_spacing(line, self.count)
            else:
                line = self._apply_spacing(line, self.count + self.extra_for_new_lines)

            first_line = False
            lines[index] = line.rstrip()

        return "\n".join(lines)


def docstring_section(title: str, title_colon: bool = True, indent: Indent = Indent()):
    """
    Builds a section of the docstring e.g. Args, Returns, etc.

    The decorated method should return a BaseDoc or list of
    BaseDoc objects.

    Args:
        title (str): the title of the section.
        title_colon (bool): suffix the title with a colon.
        indent (Indent): indentation for non-title lines.
    """

    def customized_decorator(func):
        @wraps(func)
        def decorator(self, *args, **kwargs) -> Optional[str]:
            value = func(self, *args, **kwargs)

            if not value:
                return None

            lines = []

            if title_colon:
                lines.append(f"{title}:")
            else:
                lines.append(title)

            # single BaseDoc object
            if isinstance(value, BaseDoc):
                lines.append(indent(str(value)))

            # list of BaseDoc objects
            elif isinstance(value, (tuple, list)):
                for item in value:
                    if not isinstance(item, BaseDoc):
                        raise ValueError(f"unexpected list item: {item}")

                    lines.append(indent(str(item)))

            else:
                raise ValueError(f"unexpected function result: {value}")

            return "\n".join(lines)

        return property(decorator)

    return customized_decorator


class BaseDoc:  # pylint:disable=too-few-public-methods
    """
    The base class for all docstring pieces.

    Subclasses should implement the __str__ method.
    """

    def __str__(self):
        """
        To be implemented by the subclass.
        """
        raise NotImplementedError


class Docstring(BaseDoc):  # pylint:disable=too-few-public-methods
    """
    Base class for dynamically building docstrings consisting
    of a number of sections.

    Attributes:
        _sections: list of ordered attributes corresponding to
            sections of the docstring.
    """

    _sections = []

    def _get_sections(self) -> List[str]:
        """
        Returns the list of sections for the docstring.
        """
        sections = []

        for attr in self._sections:
            value = getattr(self, attr)

            if value:
                sections.append(value)

        return sections

    def __str__(self):
        sections = self._get_sections()
        return "\n\n".join(sections)


@dataclass
class ClassAttribute(BaseDoc):
    """
    Describes a class attribute.

    Format:
    name: description

    Attributes:
        name (str): the name of the attribute.
        description (str): verbose description of the
            attribute.
    """

    name: str
    description: str = None

    def __str__(self):
        result = self.name

        if self.description:
            result += f": {self.description}"

        return result


@dataclass
class ClassDocstring(Docstring):
    """
    Constructs a doc string for a class.

    Attributes:
        summary (str): short, single line summary of the class.
        description (str): longer, verbose description of the class.
        attrs (list): list of class attributes.
    """

    summary: str = None
    description: str = None
    attrs: List[ClassAttribute] = dataclass_field(default_factory=lambda: [])

    _sections = ["fsummary", "description", "fattrs"]

    @property
    def fsummary(self):
        """Ensures summary is non-empty for formatting."""
        return self.summary or " "

    @docstring_section("Attributes")
    def fattrs(self) -> Optional[str]:
        """Formatted Attributes section"""
        return self.attrs


@dataclass
class FunctionArgument(BaseDoc):
    """
    Describes an argument to a function.

    Format:
    name (arg_type, optional): description

    Attributes:
        name (str): the name of the argument.
        arg_type (str): the type of the argument.
        optional (bool): flag if the argument is optional
            or mandatory.
        description (str): verbose description of the
            argument.
    """

    name: str
    arg_type: str = None
    optional: bool = False
    description: str = None

    def __str__(self):
        result = f"{self.name}"
        type_info = []

        if self.arg_type:
            type_info.append(self.arg_type)

        if self.optional:
            type_info.append("optional")

        if type_info:
            result += " (" + ", ".join(type_info) + ")"

        if self.description:
            result += f": {self.description}"

        return result


@dataclass
class PossibleException(BaseDoc):
    """
    Describes a possible exception that a function could raise.

    Format:
    name: description

    Attributes:
        name (str): the class name of the exception.
        description (str): verbose description of the exception
            and why it would be raised.
    """

    name: str
    description: str = None

    def __str__(self):
        result = f"{self.name}"

        if self.description:
            result += f": {self.description}"

        return result


@dataclass
class ReturnType(BaseDoc):
    """
    Describes the object returned from a function.

    Format:
    result_type: summary

    description

    Attributes:
        result_type (str): the type of the object returned.
        summary (str): short summary of the returned object.
        description (str): longer, verbose description of the
            returned object.
    """

    result_type: str
    summary: str = None
    description: str = None

    def __str__(self):
        lines = []
        first_line = self.result_type

        if self.summary:
            first_line += f": {self.summary}"

        lines.append(first_line)

        if self.description:
            lines.append("")
            lines.append(self.description)

        return "\n".join(lines)


@dataclass
class FunctionDocstring(Docstring):
    """
    Constructs a docstring for a function.

    Attributes:
        summary (str): short, single line summary of the function.
        description (str): longer, verbose description of the function.
        args (list): list of arguments for the function.
        returns (ReturnType): description of the object returned by the
            function.
        raises (list): list of possible exceptions that the function
            could raise.
    """

    summary: str = None
    description: str = None
    args: List[FunctionArgument] = dataclass_field(default_factory=lambda: [])
    returns: ReturnType = None
    raises: List[PossibleException] = dataclass_field(default_factory=lambda: [])

    _sections = ["summary", "description", "fargs", "freturns", "fraises"]

    @docstring_section("Args", indent=Indent(extra_for_new_lines=1))
    def fargs(self) -> Optional[str]:
        """Formatted Args section"""
        return self.args

    @docstring_section("Returns")
    def freturns(self) -> Optional[str]:
        """Formatted Returns section"""
        return self.returns

    @docstring_section("Raises")
    def fraises(self) -> Optional[str]:
        """Formatted Raises section"""
        return self.raises
