# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring

import inflection

from factory.base import Factory
from qctrlapiclient import Api, resourceobject
from qctrlapiclient.model import Model

from qctrl.dynamic import register_dynamic_class
from qctrl.utils import get_resource_schema, normalize_encoded_field_name
from .config import get_config
from .doc import DocGenerator
from .model_factory import ModelFactory


class ModelWrapper:  # pylint:disable=too-few-public-methods
    """Wraps a model instance so it can be used in a factory.
    """

    def __new__(cls, model_instance: Model):
        return type(
            "ModelWrapper",
            (),
            dict(__new__=lambda cls, *args, **kwargs: model_instance.new()),
        )


class FactoryBuilder:  # pylint:disable=too-few-public-methods
    """Builds factories for models.
    """

    SCHEMA_CUSTOM_DOC_KEY = "x-schema-doc"

    def __init__(self, api: Api):
        self._api = api
        self._doc = DocGenerator(api._api_schema)

    def __call__(self, resource_name: str, model) -> list:
        """Returns list of Factory classes for the model.
        """
        factories = []

        try:
            schema = self._api._model_schema[resource_name]
        except KeyError:
            raise KeyError("schema not found for {}".format(resource_name))

        BaseFactory = self._create_base(model)

        # create model factory
        model_factory = self._create(BaseFactory, resource_name, schema)
        factories.append(model_factory)

        # create typed factories (if any)
        typed_components = self._api._api_schema.get("x-typed-components")

        for entry in typed_components:
            if resource_name == entry.get("resource_type"):
                typed_model_factory = self._create(
                    model_factory,
                    resource_name,
                    schema,
                    _type=entry.get("type"),
                    restrict_fields=normalize_encoded_field_name(
                        dict_keys=entry.get("fields", {})
                    ),
                )

                factories.append(typed_model_factory)

        return factories

    @staticmethod
    def _create_base(model_instance):
        """Creates the base factory class for the model.
        """

        class BaseFactory(ModelFactory):
            """Represent BaseFactory
            """

            class Meta:  # pylint:disable=too-few-public-methods
                """Meta config for BaseFactory.
                """

                model = ModelWrapper(model_instance)

        return BaseFactory

    @register_dynamic_class("factories")
    def _create(
        self,
        base_class: Factory,
        resource_name: str,
        schema: dict,  # pylint:disable=unused-argument
        _type: str = None,
        restrict_fields: dict = None,
    ) -> Factory:
        """Creates a factory class.
        """
        factory_name = self._get_name(resource_name, _type)
        _attrs = {}

        if _type:
            _attrs.update({"type": _type})

        # add custom attrs from config
        custom_attrs = self._get_custom_attrs(factory_name)
        extra_attrs = custom_attrs.copy()
        extra_attrs.update(self._get_parent_custom_attrs(base_class))
        _attrs.update(extra_attrs)

        model_name = self._get_model_name(resource_name)

        # create new factory class
        new_factory = type(factory_name, (base_class,), _attrs)

        # get doc fields for factory
        fields = self._doc.get_fields(
            resource_name,
            only=list(restrict_fields.keys()) if restrict_fields else None,
            overrides=restrict_fields,
            write_only=True,
        )

        # add class-level documentation
        new_factory.__doc__ = f"A factory class used to create {model_name} objects."

        # new create function to be modified with documentation
        def create(cls, *args, **kwargs) -> resourceobject.ResourceObject:
            self._preprocess_kwargs(factory_name, kwargs)

            return super(new_factory, cls).create(  # pylint:disable=bad-super-call
                *args, **kwargs
            )

        # add documentation for create function
        create.__doc__ = self._doc.get_docstring(
            fields,
            "Keyword Args",
            self._get_extra_text(resource_name, factory_name, model_name),
            return_object=model_name,
        )

        # modify function signature
        # wrapper = self._doc.get_signature_wrapper(fields)
        # create = wrapper(create)

        # override default create function
        new_factory.create = classmethod(create)

        return new_factory

    @staticmethod
    def _get_name(resource_name: str, _type: str = None) -> str:
        """Returns the class name for a new factory.
        """
        name = inflection.singularize(resource_name)
        name = "{}_factory".format(name)

        if _type:
            name = "{}_{}".format(_type, name)

        name = inflection.camelize(name)
        return name

    @staticmethod
    def _get_model_name(resource_name: str) -> str:
        """Returns a model class name of a resource
        """
        name = inflection.singularize(resource_name)
        name = inflection.camelize(name)
        return name

    @classmethod
    def _get_parent_custom_attrs(cls, factory_cls):
        """
        Recursively get any custom attributes
        defined in the config for the given
        class and its parents.
        """
        attrs = {}

        if not issubclass(factory_cls, ModelFactory):
            return attrs

        for parent_cls in factory_cls.__bases__:
            attrs.update(cls._get_parent_custom_attrs(parent_cls))

        attrs.update(cls._get_custom_attrs(factory_cls.__name__))
        return attrs

    @staticmethod
    def _get_custom_attrs(factory_name: str) -> dict:
        """Returns any custom attributes defined in the config.
        """
        ref = f"FACTORIES.{factory_name}.attrs"
        return get_config(ref, {})

    @staticmethod
    def _preprocess_kwargs(factory_name: str, kwargs: dict) -> dict:
        """Updates kwargs according to any customization defined in the config.
        """
        field_mappers = get_config(f"FACTORIES.{factory_name}.preprocess", {})

        for key, value in kwargs.items():
            if key in field_mappers:
                func = field_mappers[key]
                kwargs[key] = func(value)

    def _get_extra_text(
        self, resource_name: str, factory_name: str, model_name: str
    ) -> str:
        """
        custom doc - from schema
        extra text - from config
        """

        lines = [f"Create a {model_name}."]

        # get custom doc from schema (if any)
        schema = get_resource_schema(
            self._api._api_schema, resource_name  # pylint:disable=protected-access
        )
        schema_doc = schema.get(self.SCHEMA_CUSTOM_DOC_KEY)

        if schema_doc:
            lines.append(schema_doc.strip())

        # get custom doc from config (if any)
        config_doc = get_config(f"FACTORIES.{factory_name}.doc")

        if config_doc:
            lines.append(config_doc.strip())

        return "\n\n".join(lines)


__all__ = ["FactoryBuilder"]
