# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring

import inflection

from qctrlapiclient import Api
from qctrlapiclient.model import Model

from qctrl.dynamic import register_dynamic_class
from .doc import DocGenerator


class ModelBuilder:  # pylint:disable=too-few-public-methods
    """Builds a model class.
    """

    def __init__(self, api: Api):
        self._api = api
        self._doc = DocGenerator(api._api_schema)

    def __call__(self, resource_name: str, model_instance: Model) -> type:
        """Returns a model class.
        """

        klass = self._create(resource_name, model_instance)
        return klass

    @register_dynamic_class("models")
    def _create(self, resource_name: str, model_instance: Model) -> type:
        """Creates a model class.
        """
        model_name = self._get_model_name(resource_name)
        fields = self._doc.get_fields(resource_name)

        # create new model class
        return type(
            model_name,
            (model_instance.__class__,),
            {"__doc__": self._doc.get_docstring(fields, "Attributes"),},
        )

    @staticmethod
    def _get_model_name(resource_name: str) -> str:
        """Returns the class name to be used for the model.
        """
        name = inflection.singularize(resource_name)
        return inflection.camelize(name)
