# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""A FactoryBoy implementation for models within the Q-CTRL API client package
"""

from factory.base import Factory


class ModelFactory(Factory):
    """Factory base support for sequences, attributes and stubs.
    """

    @classmethod
    def create(cls, **kwargs):
        """Create an instance of the associated class, with overriden attrs."""
        obj = super().create(**kwargs)

        # if a model has related models, refresh those model
        # to ensure the relationship is valid

        related_objects = [
            attr
            for attr in obj._relationships.keys()  # pylint:disable=protected-access
            if attr in kwargs.keys()
        ]
        for related_obj in related_objects:
            related_obj = kwargs[related_obj]
            related_obj.refresh()

        return obj

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        """Create an instance of the model_class.
        Customization point, will be called once the full set of args and kwargs
        has been computed.

        Args:
            model_class (type): the class for which an instance should be
                created
            args (tuple): arguments to use when creating the class
            kwargs (dict): keyword arguments to use when creating the class
        """
        obj = cls._build(model_class, *args, **kwargs)
        obj.commit()
        return obj

    @classmethod
    def _build(cls, model_class, *args, **kwargs):
        """Build an instance of the associated class, with overriden attrs.
        """
        obj = model_class()

        for key, value in kwargs.items():
            if key == "type":
                obj.fields[key] = value
            else:
                setattr(obj, key, value)

        return obj
