# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import inspect
import logging
import warnings

from types import SimpleNamespace
from typing import Any, Callable, Dict, List

import inflection

from qctrl.dynamic import register_dynamic_class


LOGGER = logging.getLogger(__name__)


FACTORY_NAMESPACE_DOCSTRING = """
    Namespace for factory classes. Factories are used to create
    models to be used in services.
"""


FUNCTION_NAMESPACE_DOCSTRING = """
    Namespace for functions. Functions are computations using objects
    created from the `types` namespace.
"""


MODEL_NAMESPACE_DOCSTRING = """
    Namespace for model instances. Models represent various components
    of a quantum system.

    Attributes:
        actions (qctrl.dynamic.models.Action):
        cards (qctrl.dynamic.models.Card):
        controls (qctrl.dynamic.models.Control):
        filter_functions (qctrl.dynamic.models.FilterFunction):
        graphs (qctrl.dynamic.models.Graph):
        noise_characterizations (qctrl.dynamic.models.NoiseCharacterization):
        noise_reconstructions (qctrl.dynamic.models.NoiseReconstruction):
        noise_spectral_densities (qctrl.dynamic.models.NoiseSpectralDensity):
        noises (qctrl.dynamic.models.Noise):
        plans (qctrl.dynamic.models.Plan):
        products (qctrl.dynamic.models.Product):
        pulses (qctrl.dynamic.models.Pulse):
        quasi_static_functions (qctrl.dynamic.models.QuasiStaticFunction):
        simulations (qctrl.dynamic.models.Simulation):
        subscriptions (qctrl.dynamic.models.Subscription):
        systems (qctrl.dynamic.models.System):
        targets (qctrl.dynamic.models.Target):
        trajectories (qctrl.dynamic.models.Trajectory):
        users (qctrl.dynamic.models.User):
"""


OPERATION_NAMESPACE_DOCSTRING = """
    Namespace for operations. Operations are used to build graphs for
    remote execution of complex computation.
"""


SERVICE_NAMESPACE_DOCSTRING = """
    Namespace for services. Services are computations using models.

    Attributes:
        colored_noise_simulations (qctrl.static.services.ColoredNoiseSimulationService):
        coherent_simulations (qctrl.static.services.CoherentSimulationService):
        filter_functions (qctrl.static.services.FilterFunctionService):
        relaxation_infidelity (qctrl.static.services.T1DecayAssessmentService):
        noise_free_metrics (qctrl.static.services.NoiseFreeMetricsService):
        robust_optimization (qctrl.static.services.RobustOptimizedSystemService):
        optimal_optimization (qctrl.static.services.OptimalOptimizedSystemService):
        piecewise_power_law_noise_spectral_density (qctrl.static.services.PiecewisePowerLawNoiseSpectralDensityService):
        quasi_static_functions (qctrl.static.services.QuasiStaticFunctionService):
        noise_characterization (qctrl.static.services.NoiseCharacterizationService):
        noise_reconstructions (qctrl.static.services.NoiseReconstructionService):
        flexible_optimize (qctrl.static.services.FlexibleOptimizedSystemService):
"""


TYPE_NAMESPACE_DOCSTRING = """
    Namespace for types. Objects created from these types are used
    when performing computations defined in the `functions` namespace.
"""


class BaseNamespace(SimpleNamespace):
    """
    Base namespace class for Qctrl components.
    """

    @classmethod
    def extend(cls, attrs: Dict[str, Any]):
        """
        Extends the namespace class by adding attributes.

        Args:
            attrs (dict): dict of attribues to add to the namespace.

        Raises:
            AttributeError: existing attribute found.
        """
        for attr, value in attrs.items():
            if hasattr(cls, attr):
                raise AttributeError(f"existing attr ({attr}) on namespace: {cls}")

            LOGGER.debug("adding attr %s to namespace: %s", attr, cls)
            setattr(cls, attr, value)

    @classmethod
    def extend_functions(cls, *funcs: Callable):
        """
        Extends the namespace class by adding functions
        as attributes. The function will be added as a
        staticmethod.

        Args:
            funcs: functions to be added to the namespace.
        """
        for func in funcs:
            cls.extend({func.__name__: staticmethod(func)})


class FactoryNamespaceMixin:  # pylint:disable=too-few-public-methods
    """
    Legacy mixin for factory namespace. Mimics `klass_registry` functionality.
    """

    @classmethod
    def get_class(cls, name: str) -> type:
        """
        Returns the corresponding Factory class for the given name.

        Args:
            name (str): the name of the factory class e.g. NoiseFactory

        Returns:
            type: the Factory class

        Raises:
            ValueError: Factory class not found
        """
        warnings.warn(
            "The `get_class` method is deprecated. Access the factory using the named attribute."
        )

        for value in cls.__dict__.values():
            if inspect.isclass(value) and value.__name__ == name:
                return value

        raise ValueError(f"factory class not found: {name}")


@register_dynamic_class("namespaces")
def _create_namespace_cls(
    prefix: str,
    docstring: str = None,
    base: SimpleNamespace = BaseNamespace,
    mixins: List[type] = None,
) -> SimpleNamespace:
    """
    Creates a new namespace class and returns an instance of it.
    """
    name = f"{inflection.camelize(prefix)}Namespace"
    parents = [base]

    if mixins:
        parents += mixins

    attrs = {}

    if docstring:
        attrs.update({"__doc__": docstring})

    new_cls = type(name, tuple(parents), attrs)
    return new_cls


def create_factory_namespace():
    """
    Creates the factory namespace.
    """
    cls = _create_namespace_cls(
        "Factory", FACTORY_NAMESPACE_DOCSTRING, mixins=[FactoryNamespaceMixin]
    )
    return cls()


def create_function_namespace():
    """
    Creates the function namespace.
    """
    cls = _create_namespace_cls("Function", FUNCTION_NAMESPACE_DOCSTRING)
    return cls()


def create_model_namespace():
    """
    Creates the model namespace.
    """
    cls = _create_namespace_cls("Model", MODEL_NAMESPACE_DOCSTRING)
    return cls()


def create_operation_namespace():
    """
    Creates the operation namespace.
    """
    cls = _create_namespace_cls("Operation", OPERATION_NAMESPACE_DOCSTRING)
    return cls()


def create_service_namespace():
    """
    Creates the service namespace.
    """
    cls = _create_namespace_cls("Service", SERVICE_NAMESPACE_DOCSTRING)
    return cls()


def create_type_namespace():
    """
    Creates the type namespace.
    """
    cls = _create_namespace_cls("Type", TYPE_NAMESPACE_DOCSTRING)
    return cls()


__all__ = [
    "BaseNamespace",
    "create_factory_namespace",
    "create_function_namespace",
    "create_model_namespace",
    "create_operation_namespace",
    "create_service_namespace",
    "create_type_namespace",
]
