# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import json
import logging

from collections import OrderedDict
from typing import Any, Dict, List, Union
from graphql import (
    GraphQLArgument,
    GraphQLNonNull,
    GraphQLInt,
    GraphQLField,
    GraphQLFloat,
    GraphQLString,
    GraphQLBoolean,
    GraphQLScalarType,
    GraphQLInputObjectField,
    GraphQLList,
    GraphQLInputObjectType,
    GraphQLEnumType,
)
from gql import gql
from gql_query_builder import GqlQuery
from qctrlcommons.serializers import DataTypeEncoder

from qctrl.utils import _is_private_field


LOGGER = logging.getLogger(__name__)


class QueryBuilder:  # pylint:disable=too-few-public-methods
    """
    Translates function args/kwargs to GraphQL queries.
    """

    def __init__(self, name: str, mutation_field: GraphQLField):
        self._name = name
        self._input_fields = self._build_input_fields(mutation_field)
        self._output_fields = self._build_output_fields(mutation_field)

    @classmethod
    def _build_input_fields(cls, mutation_field: GraphQLField) -> OrderedDict:
        """
        Given the GraphQL Mutation field, return an OrderedDict
        of sub-fields which comprise of the input field. Assumes
        the mutation field has a single argument, named `input`.
        The ordering of fields is maintained from the schema.
        """
        try:
            input_arg = mutation_field.args["input"]
        except KeyError:
            raise KeyError(f"mutation field does not have input arg: {mutation_field}")

        if not isinstance(input_arg, GraphQLArgument):
            raise ValueError(f"unexpected input arg type: {input_arg}")

        field_type = input_arg.type

        # the input arg type should always non-null
        if isinstance(field_type, GraphQLNonNull):
            field_type = field_type.of_type

        input_fields = OrderedDict()

        # remove ignored fields. note: field_type.fields
        # is ordered. see: graphql.pyutils.ordereddict
        for key, field in field_type.fields.items():
            if _is_private_field(key):
                continue

            input_fields[key] = field

        return input_fields

    @classmethod
    def _build_output_fields(cls, field: GraphQLField) -> List[str]:
        """
        Recursively builds the list of output field names
        for the given GraphQLField.
        """
        LOGGER.info("building output fields for %s", field)
        field_type = field.type

        # get wrapped type
        if isinstance(field_type, GraphQLNonNull):
            field_type = field_type.of_type

        # no nested fields
        if isinstance(field_type, (GraphQLScalarType, GraphQLEnumType, GraphQLList)):
            return None

        result = []

        for name, sub_field in field_type.fields.items():
            sub_result = cls._build_output_fields(sub_field)

            # no nested fields, just select field name
            if sub_result is None:
                result.append(name)

            # has nested fields, include nested field query and field name
            else:
                sub_result = GqlQuery().fields(sub_result).generate()
                sub_result = f"{name} {sub_result}"
                LOGGER.debug("building nested output query: %s", sub_result)
                result.append(sub_result)

        return result

    def get_input_fields(self) -> OrderedDict:
        """
        Returns the ordered input fields for the query.
        """
        return self._input_fields

    def _build_input_values(
        self, args: List[Any], kwargs: Dict[str, Any]
    ) -> OrderedDict:
        """
        Given the args and kwargs passed to the function call,
        return an OrderedDict of values for each input field,
        if provided.
        """
        input_values = OrderedDict()

        # get the input field names in order for processing args
        field_names = list(self._input_fields.keys())

        # for each positional arg
        for index, value in enumerate(args):

            # check end bound
            if index >= len(field_names):
                raise IndexError("too many args provided")

            # store value with corresponding field name
            field_name = field_names[index]
            input_values[field_name] = value

        # for each kwarg
        for key, value in kwargs.items():

            # this is likely to be a default value
            if value is None:
                continue

            # duplicate value - this shouldnt happen
            if key in input_values:
                raise KeyError(f"duplicate argument: {key}")

            # unknown arg - this shouldnt happen
            if key not in self._input_fields:
                raise KeyError(f"unknown argument: {key}")

            input_values[key] = value

        return input_values

    @staticmethod
    def _get_nested_object_value(value: Any, key: str) -> Any:
        """
        Returns the nested value at the key. If value is
        a dictionary, returns the corresponding key value.
        Otherwise, assumes the nested value is at the `key`
        attribute of value.
        """
        if isinstance(value, dict):
            return value[key]

        return getattr(value, key)

    @classmethod
    def _format_field_value(
        cls,
        field: Union[GraphQLInputObjectField, GraphQLInputObjectType, Dict],
        value: Any,
    ):
        """
        Recursively formats a value for use in a GraphQL query.
        """
        LOGGER.debug("formatting - field:%s, value:%s", field, value)

        # input field wrapper
        if isinstance(field, GraphQLInputObjectField):
            return cls._format_field_value(field.type, value)

        # non null
        if isinstance(field, GraphQLNonNull):
            return cls._format_field_value(field.of_type, value)

        # list
        if isinstance(field, GraphQLList):
            nested_formatted_values = []
            LOGGER.debug("formatting list of type %s", field.of_type)

            # format each value in the list
            for nested_value in value:
                formatted_value = cls._format_field_value(field.of_type, nested_value)
                nested_formatted_values.append(formatted_value)

            return "[ {} ]".format(", ".join(nested_formatted_values))

        # custom object
        if isinstance(field, GraphQLInputObjectType):
            # process the dict of sub-fields
            return cls._format_field_value(field.fields, value)

        # dict (should be ordered - see graphql.pyutils.ordereddict)
        if isinstance(field, dict):
            LOGGER.debug("formatting ordered dict")
            nested_formatted_values = []

            # format each nested value in the odict
            for key, nested_field in field.items():
                LOGGER.debug("nested field: %s", nested_field)

                try:
                    nested_value = cls._get_nested_object_value(value, key)
                except (KeyError, AttributeError):
                    LOGGER.debug("nested object field '%s' missing - skipping", key)
                    continue

                LOGGER.debug("nested value: %s", nested_value)
                formatted_value = cls._format_field_value(nested_field, nested_value)
                nested_formatted_values.append(f"{key} : {formatted_value}")

            result = GqlQuery().fields(nested_formatted_values).generate()
            return result

        # scalar
        if isinstance(field, GraphQLScalarType):
            result = json.dumps(value, cls=DataTypeEncoder)

            # custom scalars e.g. numpy arrays, complex numbers, etc.
            # need to be "double-encoded" so that the value looks
            # like a string literal
            if cls._is_custom_scalar(field):
                result = json.dumps(result)

            return result

        # unhandled field
        raise ValueError(f"unhandled field type: {field}")

    @staticmethod
    def _is_custom_scalar(field: GraphQLInputObjectType) -> bool:
        """
        Checks if the field is a custom scalar type.
        """
        return field.name not in (
            GraphQLBoolean.name,
            GraphQLFloat.name,
            GraphQLInt.name,
            GraphQLString.name,
        )

    def __call__(self, args: List[Any], kwargs: Dict[str, Any]):
        """
        Builds a GraphQL query based on the inputs passed to the function.
        """
        # prepare inputs
        input_values = self._build_input_values(args, kwargs)
        LOGGER.debug("input_values:%s", input_values)
        query_input = self._format_field_value(self._input_fields, input_values)
        LOGGER.debug("query_input:%s", query_input)

        # build and compile query
        query = (
            GqlQuery()
            .query(self._name, input={"input": query_input})
            .fields(self._output_fields)
            .operation("mutation", name="create_action")
            .generate()
        )
        LOGGER.debug("query:%s", query)
        return gql(query)
