# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging

from collections import OrderedDict
from typing import Callable, Union

import forge

from graphql import (
    GraphQLNonNull,
    GraphQLInputObjectField,
)
from graphql.type.definition import GraphQLType
from .type_registry import TypeRegistry


LOGGER = logging.getLogger(__name__)


class SignatureBuilder:  # pylint:disable=too-few-public-methods
    """
    Builds function signatures from GraphQL fields.
    """

    def __init__(self, input_fields: OrderedDict, type_registry: TypeRegistry = None):
        self._type_registry = type_registry
        self._signature = self._build(input_fields)

    def _get_type_hint(
        self, field_type: GraphQLType
    ) -> Union[type, forge.FParameter.empty]:
        """
        Returns a type hint for the GraphQL field type.
        If no type hint can be determined, forge.FParameter.empty
        is returned.
        """
        type_hint = None

        if self._type_registry:
            type_hint = self._type_registry.get_type_hint(field_type)

        return type_hint or forge.FParameter.empty

    @staticmethod
    def _is_kwarg(field: GraphQLInputObjectField) -> bool:
        """
        Checks if the field corresponds to a
        keyword arg for a function signature.
        """
        # field has a default value
        if field.default_value is not None:
            return True

        # field is optional
        if not isinstance(field.type, GraphQLNonNull):
            return True

        # otherwise, arg is positional (depending
        # on context)
        return False

    def _build(self, input_fields: OrderedDict):
        """
        Returns a function signature based on the input fields.
        Assumes that the fields are organised such that positional
        args (required fields) are followed by keyword args (optional
        fields and/or with default values).
        """

        args = []

        # flag to indicate if future args should be kwargs
        using_kwargs = False

        for name, field in input_fields.items():

            # check for required field after optional field
            # which would correspond to positional arg after
            # kwarg
            if using_kwargs and isinstance(field.type, GraphQLNonNull):
                raise RuntimeError(f"required field '{name}' follows optional field")

            # if processing positional args and we encounter
            # a keyword arg, all future args should be keyword
            if not using_kwargs and self._is_kwarg(field):
                using_kwargs = True

            # use default default value
            default_value = forge.FParameter.empty

            # set default value for kwarg - even if the default
            # is None
            if using_kwargs:
                default_value = field.default_value

            # build arg and add to list
            args.append(
                forge.arg(
                    name, type=self._get_type_hint(field.type), default=default_value,
                )
            )

        LOGGER.debug("signature args:%s", args)

        # build signature
        return forge.sign(*args)

    def __call__(self, func: Callable):
        """
        Signs the function with the signature.
        """
        return self._signature(func)
