# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import dataclasses
import logging

from enum import Enum
from typing import Any, Dict, Optional, List, Tuple

import numpy as np

from graphql import (
    GraphQLNonNull,
    GraphQLScalarType,
    GraphQLList,
    GraphQLInputObjectField,
    GraphQLEnumType,
)
from graphql.type.definition import GraphQLType

from qctrl.dynamic import register_dynamic_class
from qctrl.utils import _is_private_field
from .doc import ClassAttribute, ClassDocstring


LOGGER = logging.getLogger(__name__)


class TypeRegistry:
    """
    Builds dataclasses from GraphQL types.
    """

    SCALAR_TYPES = {
        "String": str,
        "Boolean": bool,
        "Int": int,
        "Float": float,
        "ID": str,
        # custom
        "EncodedComplexNumber": complex,
        "EncodedNumpyArray": np.ndarray,
    }

    def __init__(self):
        self._types = {}

    def get_type_hint(self, obj: GraphQLType) -> type:
        """
        Returns the type hint for the GraphQL type.
        If there is no corresponding type, a new
        type is created.
        """
        LOGGER.debug("getting type hint for: %s", obj)

        # non-null
        if isinstance(obj, GraphQLNonNull):
            return self.get_type_hint(obj.of_type)

        # scalar
        if isinstance(obj, GraphQLScalarType):
            if obj.name in self.SCALAR_TYPES:
                return self.SCALAR_TYPES[obj.name]

            LOGGER.warning("unhandled scalar type: %s", obj)
            return None

        # list
        if isinstance(obj, GraphQLList):
            return List[self.get_type_hint(obj.of_type)]

        # existing type
        if obj.name in self._types:
            return self._types[obj.name]

        # build new type
        return self.build(obj)

    def get_type_map(self) -> Dict[str, type]:
        """
        Returns the type map.
        """
        return self._types

    def build(self, obj: GraphQLType) -> type:
        """
        Builds a new data class from the
        GraphQL type.
        """
        LOGGER.debug("building type for %s", obj)

        # handle non-null
        if isinstance(obj, GraphQLNonNull):
            return self.build(obj.of_type)

        # handle enum
        if isinstance(obj, GraphQLEnumType):
            values = {}

            for enum_value in obj.values:
                values[enum_value.name] = enum_value.value

            docstring = ClassDocstring(description=obj.description)
            return self._create_enum(obj.name, values, docstring)

        # handle complex type with sub fields
        fields = []
        docstring = ClassDocstring(description=obj.description)

        # for each sub field
        for field_name, field in obj.fields.items():

            # ignore private fields
            if not _is_private_field(field_name):

                # get type hint for sub field
                type_hint = self.get_type_hint(field.type)

                # sub field has default value
                if (
                    isinstance(field, GraphQLInputObjectField)
                    and field.default_value is not None
                ):
                    fields.append(
                        (
                            field_name,
                            type_hint,
                            dataclasses.field(default=field.default_value),
                        )
                    )

                # sub field has no default value
                else:
                    fields.append((field_name, type_hint))

                # add attr for docstring
                docstring.attrs.append(ClassAttribute(field_name, field.description))

        # create new dataclass
        cls = self._create_dataclass(obj.name, fields, docstring=str(docstring))

        return cls

    @register_dynamic_class("types")
    def _create_enum(
        self, name: str, values: Dict[str, Any], docstring: str = None
    ) -> type:
        """
        Creates a new enum class and registers it.
        """
        cls = Enum(name, values)

        if docstring:
            cls.__doc__ = str(docstring)

        self._register(cls)
        return cls

    @register_dynamic_class("types")
    def _create_dataclass(
        self, name: str, fields: List[Tuple[str, Optional[str]]], docstring: str = None,
    ) -> type:
        """
        Creates a new data class and registers it.
        """
        cls = dataclasses.make_dataclass(name, fields)

        if docstring:
            cls.__doc__ = str(docstring)

        self._register(cls)
        return cls

    def _register(self, cls: type):
        """
        Registers the new type class.
        """
        key = cls.__name__

        if key in self._types:
            raise KeyError(f"duplicate type name: {key}")

        self._types[key] = cls
