"""
Dynamically created classes.
"""
import inspect
import sys

from functools import wraps
from typing import Callable

# these imports setup the sub modules
from .factories import *
from .models import *
from .namespaces import *
from .types import *


def register_dynamic_class(submodule: str):
    """
    Decorator to register a dynamic class. The
    decorated function should return the dynamic
    class to be registered.

    Args:
        submodule (str): the submodule name that
            the class should be registered under
            e.g. `types`
    """

    def customized_decorator(func: Callable):
        @wraps(func)
        def decorator(*args, **kwargs):
            cls = func(*args, **kwargs)
            assert inspect.isclass(cls)
            module = f"qctrl.dynamic.{submodule}"
            cls.__module__ = module
            cls.__package__ = module
            setattr(sys.modules[module], cls.__name__, cls)
            return cls

        return decorator

    return customized_decorator


__all__ = ["register_dynamic_class"]
