# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import json
import logging
import os
from pathlib import Path
from typing import Dict
from urllib.parse import urljoin

import jsonschema
import pythonflow as pf
import tenacity
from dateutil.parser import isoparse
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
from jsonapi_client.exceptions import DocumentError
from qctrlapiclient import api
from qctrlapiclient.exceptions import QctrlApiException
from qctrlcommons.utils import generate_user_agent
from requests.exceptions import BaseHTTPError, RequestException

from . import __version__ as qctrl_version
from .builders import build_namespaces
from .globals import _APP_CONTEXT_STACK, API_VERSION
from .scripts import file_path_for_url, interactive_authentication, write_auth_file

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods, unnecessary-lambda, too-many-instance-attributes
class Qctrl:
    """A mediator class. Used to authenticate with Q-CTRL and access Q-CTRL features.

    Creating an instance of this class requires authentication with Q-CTRL's API.

    The recommended method of authentication is through the interactive authentication
    method. This method can be invoked by simply calling Qctrl() without any arguments.
    This method will also create an authentication file that will be used for subsequent
    authentications when using the package.

    q = Qctrl()

    If needed authentication can also be done by passing your email and password as arguments
    to the Qctrl() function as shown below. Ensure that the credentials used are secure.

    q = Qctrl(email='myemail', password='mypassword')

    Legacy authentication using a single auth token is still supported for users that have such
    a token:

    q = Qctrl(auth_token="...")

    However users should change to the recommended authentication method as soon as possible.

    Args:
        auth_token (str, optional): A Q-CTRL authentication token.
        email (str, optional): The email address for a Q-CTRL account.
        password (str, optional): The password for a Q-CTRL account.
        api_root (str, optional): The URL of the Q-CTRL API.

    Raises:
        QctrlApiException

    Attributes:
        factories (qctrl.dynamic.namespaces.FactoryNamespace):
        functions (qctrl.dynamic.namespaces.FunctionNamespace):
        models (qctrl.dynamic.namespaces.ModelNamespace):
        operations (qctrl.dynamic.namespaces.OperationNamespace):
        services (qctrl.dynamic.namespaces.ServiceNamespace):
        types (qctrl.dynamic.namespaces.TypeNamespace):
    """

    _credentials_file_schema = {
        "type": "object",
        "properties": {"email": {"type": "string"}, "password": {"type": "string"}},
        "required": ["email", "password"],
    }
    _credentials_file_permissions = "600"

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=5),
        stop=tenacity.stop_after_attempt(3),
        retry=tenacity.retry_if_exception_type((RequestException, BaseHTTPError)),
    )
    def __init__(
        self,
        auth_token: str = None,
        email: str = None,
        password: str = None,
        api_root: str = None,
    ):
        self._api_root = api_root or os.environ.get("QCTRL_API_HOST", api.API_ROOT)

        self.api = self._create_api(auth_token, email, password)
        self._validate_auth()
        self._gql_client = self._create_gql_client()

        # add namespaces
        (
            self.factories,
            self.functions,
            self.models,
            self.operations,
            self.services,
            self.types,
        ) = build_namespaces(self.api, self._gql_client)

        # register as current app
        _APP_CONTEXT_STACK.push(self)

    @property
    def gql_client(self) -> Client:
        """Access graphql client from Qctrl object.
        """
        return self._gql_client

    def _get_jwt_api(
        self,
        api_headers: Dict[str, str] = None,
        email: str = None,
        password: str = None,
    ) -> api.Api:
        """
        Creates an Api object using JWT authentication.

        Args:
            api_headers (dict, optional): any additional HTTP headers required by the API
            email (str, optional): user`s email address
            password (str, optional): user`s password

        Returns:
            api.Api: an Api object which uses JWT authentication.

        Raises:
            RuntimeError: missing credentials e.g. provided email but no password
        """
        token_observer = None

        # via args
        if email or password:

            if email is None:
                raise RuntimeError("email is required.")

            if password is None:
                raise RuntimeError("password is required.")

            access_token, refresh_token = api.Api.login(
                self._api_root, email, password, api_version=API_VERSION
            )

        # via file
        else:

            file_path = self._credentials_file_path
            access_token, refresh_token = self._parse_credentials_file(file_path)

            token_observer = lambda jwt: write_auth_file(
                jwt["access_token"], jwt["refresh_token"], file_path=file_path
            )

        return api.Api(
            access_token,
            refresh_token=refresh_token,
            api_root=self._api_root,
            headers=api_headers,
            token_observer=token_observer,
            api_version=API_VERSION,
        )

    def _create_api(
        self, auth_token: str = None, email: str = None, password: str = None
    ) -> api.Api:
        """
        Creates an Api object for accessing the server. If auth_token is provided,
        the legacy authentication method will be used; otherwise, JWT authentication
        will be used.

        Args:
            auth_token (str, optional): legacy authentication token
            email (str, optional): user`s email address
            password (str, optional): user`s password

        Returns:
            api.Api: an Api object used for communication with the server.
        """
        api_headers = {
            "User-Agent": generate_user_agent(
                "Q-CTRL Python Package", version=qctrl_version
            )
        }

        # auth via token
        if auth_token:
            return api.Api(
                auth_token,
                api_root=self._api_root,
                headers=api_headers,
                api_version=API_VERSION,
            )

        # auth via JWT
        return self._get_jwt_api(api_headers, email, password)

    def _create_gql_client(self):
        url = urljoin(self._api_root, "graphql/")
        LOGGER.debug("gql url: %s", url)

        # pylint:disable=protected-access,fixme
        auth = self.api._session._request_kwargs[
            # TODO fix reference
            "auth"
        ]
        transport = RequestsHTTPTransport(url=url, use_json=True, auth=auth,)

        return Client(retries=3, transport=transport, fetch_schema_from_transport=True)

    def _validate_auth(self):
        """Checks that the configured authentication is valid.Raises a QctrlApiException if invalid.
        """
        try:
            self.api.systems.find("first")
        except DocumentError as error:

            # unauthorized response
            if error.response.status_code == 401:  # pylint:disable=no-member
                raise QctrlApiException("Invalid token")

    @classmethod
    def _parse_credentials_file(cls, file_path: str) -> tuple:
        """Parses a credentials file and returns a tuple of
        (access_token, refresh_token) to be used for login.
        """
        if not os.path.isfile(file_path):
            raise OSError(f"credentials file does not exist: {file_path}")

        # check permissions
        stat = os.stat(file_path)
        permissions = oct(stat.st_mode)

        if not permissions.endswith(cls._credentials_file_permissions):
            raise OSError("incorrect permissions for credentials file")

        # get data
        with open(file_path) as json_file:
            data = json_file.read()
            data = json.loads(data)

        # validate data
        jsonschema.validate(cls._credentials_file_schema, data)
        return data["access_token"], data["refresh_token"]

    @property
    def _credentials_file_path(self):
        env_file_path = os.environ.get("QCTRL_AUTHENTICATION_CREDENTIALS", "")

        if env_file_path:
            return Path(env_file_path)

        default_file_path = file_path_for_url(self._api_root)
        if not default_file_path.exists():
            interactive_authentication(api_root=self._api_root)

        return default_file_path

    @staticmethod
    def create_graph():
        """Create pythonflow graph object.
        """
        return pf.Graph()

    def activity_monitor(self, limit: int = 5, offset: int = 0):
        """
            Prints a list of previously run actions to the console
            and their statuses. Allows users to filter the amount of
            actions shown as well as provide an offset.

            Args:
                limit (int): The number of previously ran actions to show.(Default is 5)
                offset (int): Offset the list of actions by a certain amount.
        """

        query = gql(
            """
            query getActions($limit: Int, $offset: Int){
                actions(limit:$limit, offset:$offset){
                    name,
                    status,
                    progress,
                    created_at,
                    updated_at
                }
            }
            """
        )

        params = {"limit": limit, "offset": offset}
        results = self._gql_client.execute(query, variable_values=params)

        # Builds output string. Formatted in a table like manner.
        actions = f'{"Name":<40} {"Status":<10} {"Progress":<10} {"Created at":<24} {"Completed at":<24}\n'  # pylint:disable=line-too-long

        for result in results["actions"]:
            created_at = "{:%d-%m-%Y %H:%M:%S}".format(isoparse(result["created_at"]))
            completed_at = "{:%d-%m-%Y %H:%M:%S}".format(isoparse(result["updated_at"]))
            action = f'\n{result["name"]:<40} {result["status"]:<10} {result["progress"]:<10.0%} {created_at:<24} {completed_at:<24}\n'  # pylint:disable=line-too-long
            actions = actions + action

        print(actions)


__all__ = [
    "Qctrl",
]
