# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import json
import logging
import time

from gql import gql, Client
from qctrlcommons.serializers import DataTypeDecoder
from tqdm import tqdm

from qctrl.globals import CURRENT_APP

LOGGER = logging.getLogger(__name__)


class Result:
    """
    Wraps the result of a GraphQL mutation.
    """

    # action statuses
    PENDING = "PENDING"
    STARTED = "STARTED"
    RETRY = "RETRY"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"

    def __init__(self, raw: dict):
        self._raw = raw

    def wait_for_completion(self, check_every_seconds: int = 2):
        """
        Waits until the corresponding action is completed on the server.
        """
        with tqdm(total=100) as progress_bar:
            _total = 0

            while not self.is_completed:
                progress = int(self._progress * 100) - _total
                _total += progress
                progress_bar.update(progress)
                time.sleep(check_every_seconds)
                self.refresh()

            progress_bar.update(max(0, progress_bar.total - progress_bar.n))

        # action is now completed

        if not self.is_successful:
            if self.errors:
                raise RuntimeError(f"Execution failed: {self.errors}")

            if self.status == self.FAILURE:
                raise RuntimeError(
                    "Execution resulted with status 'FAILURE'("
                    "no extra details about the errors are "
                    "available)."
                )

        LOGGER.info(
            "Action %s finished with status: %s", str(self.action_id), self.status
        )

    def refresh(self):
        """
        Updates the result from the server.
        """
        query = gql(
            """
            query ActionQuery($model_id: Int!) {
              action(model_id: $model_id) {
                status
                runtime
                progress
                errors {traceback}
                ... on CoreAction {
                    result
                 }
              }
            }
            """
        )
        params = {"model_id": int(self.action_id)}

        gql_client: Client = CURRENT_APP.gql_client
        query_result: dict = gql_client.execute(query, variable_values=params)
        action_result: dict = query_result["action"]

        self._raw.update(action_result)

    @property
    def is_completed(self) -> bool:  # pylint:disable=missing-function-docstring
        return self.status not in (self.PENDING, self.STARTED, self.RETRY)

    @property
    def is_successful(self) -> bool:  # pylint:disable=missing-function-docstring
        return self.status == self.SUCCESS

    @property
    def action_id(self):  # pylint:disable=missing-function-docstring
        return self._raw["model_id"]

    @property
    def name(self):  # pylint:disable=missing-function-docstring
        return self._raw["name"]

    @property
    def errors(self):  # pylint:disable=missing-function-docstring
        return {}  # TODO read from _raw pylint:disable=fixme

    @property
    def status(self):  # pylint:disable=missing-function-docstring
        return self._raw["status"]

    @property
    def _progress(self):  # pylint:disable=missing-function-docstring
        return 0  # TODO read from _raw pylint:disable=fixme

    @property
    def value(self):  # pylint:disable=missing-function-docstring
        if not self.is_completed:
            raise ValueError("Result is not ready yet. Use wait_for_completion.")

        if not self.is_successful:
            raise ValueError("Remote function failed. Result is not available.")

        return json.loads(self._raw["result"], cls=DataTypeDecoder)
