# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""Module that defines CLI scripts to be used from installed pip package.

For example:
    $ qctrl auth
"""
import hashlib
import json
import logging
import os
from pathlib import Path

import click
from qctrlapiclient import api
from qctrlcommons.exceptions import QctrlException
from requests import exceptions

from qctrl import globals as qctrl_globals

DEFAULT_AUTH_DIR = Path.home() / ".config" / "qctrl"
LOGGER = logging.getLogger(__name__)


def file_path_for_url(url: str) -> Path:
    """Returns a Path() instance for the default file location using the URL
    MD5 as the filename.
    """
    file_name = hashlib.md5(url.encode()).hexdigest()
    file_path = DEFAULT_AUTH_DIR / file_name
    return file_path


def write_auth_file(access_token: str, refresh_token: str, file_path: Path):
    """Writes the authentication file to the specified Path.
    """

    credentials = {
        "access_token": access_token,
        "refresh_token": refresh_token,
    }
    try:
        file_path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
        file_path.touch(mode=0o600, exist_ok=True)
        file_path.write_text(json.dumps(credentials))
    except IOError as exc:
        LOGGER.error("%s", exc, exc_info=True)
        raise QctrlException("incorrect permissions for credentials file") from exc


@click.group()
def main():
    """Q-CTRL CLI tool."""


@main.command()
@click.option("--access-token", required=True, help="JWT Access Token.")
@click.option("--refresh-token", required=True, help="JWT Refresh Token.")
@click.option(
    "--api-root",
    default=api.API_ROOT,
    help="Custom Q-CTRL API base URL.",
    show_default=f"{api.API_ROOT}",
)
@click.option(
    "--path",
    type=Path,
    default=lambda: os.environ.get("QCTRL_AUTHENTICATION_CREDENTIALS"),
    help=(
        "Use this option to set a custom location for the authentication files. "
        "If preset, will use the ENV variable `QCTRL_AUTHENTICATION_CREDENTIALS`. "
        "Otherwise, defaults to the standard config location."
    ),
)
def generate_auth_file(access_token, refresh_token, api_root, path: Path = None):
    """Generates the Q-CTRL Authentication file from command line."""
    if not path:
        path = file_path_for_url(api_root)

    write_auth_file(access_token, refresh_token, path)


def interactive_authentication(email=None, password=None, api_root=None, path=None):
    """Generates the authentication file interactively.
    """
    description = """
    ----------------------------------------------------------
    This is an interactive Q-CTRL Authentication setup tool.

    For non-interactive or alternative options check our help:

        $ qctrl auth --help

    ----------------------------------------------------------
    """
    if not email or not password:
        click.secho(description, fg="bright_blue")

    if not email:
        email = click.prompt("Email")
    else:
        click.secho(f"Email: {email}")

    if not password:
        password = click.prompt("Password", hide_input=True)
    else:
        click.secho("Password: <hidden>")

    if not path:
        path = file_path_for_url(api_root)

    try:
        click.secho(f"Authenticating to {api_root}")
        access_token, refresh_token = api.Api.login(
            email=email,
            password=password,
            api_root=api_root,
            api_version=qctrl_globals.API_VERSION,
        )
        click.secho("Successfully authenticated!", fg="green")
        write_auth_file(access_token, refresh_token, path)
        click.secho(f"Authentication file created at {path}")

    except exceptions.HTTPError as exc:
        LOGGER.error("%s", exc, exc_info=True)
        raise QctrlException(f"{exc}") from exc

    except (exceptions.ConnectionError, exceptions.MissingSchema) as exc:
        LOGGER.error("%s", exc, exc_info=True)
        raise QctrlException(f"{api_root} is not a valid Q-CTRL URL.") from exc


@main.command()
@click.option("--email", help="User email.")
@click.option("--password", help="User password.")
@click.option(
    "--api-root",
    default=api.API_ROOT,
    help="Custom Q-CTRL API base URL.",
    show_default=f"{api.API_ROOT}",
)
@click.option(
    "--path",
    type=Path,
    default=lambda: os.environ.get("QCTRL_AUTHENTICATION_CREDENTIALS"),
    help=(
        "Use this option to set a custom location for the authentication files. "
        "If preset, will use the ENV variable `QCTRL_AUTHENTICATION_CREDENTIALS`. "
        "Otherwise, defaults to the standard config location."
    ),
)
def auth(email, password, api_root, path=None):
    """Q-CTRL Authentication setup.

    Provides default configuration for environmental authentication,
    allowing users to avoid managing in-script credentials. After this setup,
    the Qctrl package can be used as below:

        \b
        >>> from qctrl import Qctrl
        >>> qctrl = Qctrl()

    For instance based authentication, preferred for remote deployments or
    non-interactive environments, use the standard authentication method.

        \b
        >>> from qctrl import Qctrl
        >>> qctrl = Qctrl(email=..., password=...)

    For more details, access our documentation at https://docs.q-ctrl.com
    """

    interactive_authentication(email, password, api_root, path)
