# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
import textwrap
import time
from abc import ABC

import tenacity
from jsonapi_client.exceptions import DocumentError
from qctrlapiclient import api as qctrlapi, resourceobject
from requests.exceptions import (
    ConnectionError as RequestsConnectionError,
    ConnectTimeout,
)
from tqdm import tqdm

LOGGER = logging.getLogger(__name__)


# apply to all service classes
# pylint:disable=too-few-public-methods


class BaseService(ABC):
    """A base class for interfacing with Q-CTRL services.
    """

    _label = None
    _action_type = None

    def __init__(self, api: qctrlapi.Api):
        """
        Args:
            api (qctrl.Api): Source of data
        """
        self.api = api

    def _run(
        self, resource: resourceobject.ResourceObject, name: str = None, **job_arguments
    ) -> resourceobject.ResourceObject:
        """Perform the desired action.

        Args:
          resource: The object to operate the action upon
          name: The name of the action

        Returns:
            qctrlapiclient.resourceobject.ResourceObject: The supplied resource
            updated with the result. # pylint:disable=line-too-long
        """
        # creating the action will start the action
        action = self.api.actions.new()
        action.name = textwrap.shorten(
            name or (f"{resource.__name__} action"), 50
        )  # pylint:disable=protected-access
        action.fields[
            "type"
        ] = self.__class__._action_type  # pylint:disable=protected-access
        action.related_model = resource

        if job_arguments:
            action.job_arguments = job_arguments

        action.commit()

        self._wait_for_results(action)
        resource.refresh()
        return resource

    @staticmethod
    def _wait_for_results(
        action: resourceobject.ResourceObject, check_every_seconds: float = 2
    ) -> resourceobject.ResourceObject:
        """Polling callback which will query the status of the
        running optimization.

        Args:
          action: the `Action` associated with the `System`, it will be
            a pointer to a remote running task.
          check_every_seconds: how often to check on the action for updates.
            (Defaults to 2s)

        Returns:
          ResourceObject: An instance of Action

        """

        # Note that we use tqdm as a context manager to ensure that the progress
        # bar is eventually closed (even if an exception occurs). Failing to
        # close the progress bar can cause state to leak to subsequent progress
        # bars, which leads to UI issues.
        #
        # Using `100` as total units to update the bar according to percentage
        # of subtasks completed.

        with tqdm(total=100) as progress_bar:
            _total = 0
            while action.status in ("PENDING", "STARTED", "RETRY"):
                if action.progress is not None:
                    progress = int(action.progress * 100) - _total
                    _total += progress
                    progress_bar.update(progress)
                time.sleep(check_every_seconds)
                BaseService.refresh_action(action)

            progress_bar.update(max(0, progress_bar.total - progress_bar.n))

        if action.status != "SUCCESS":
            if action.errors:
                raise RuntimeError(f"Execution failed: {action.errors}")
            if action.status == "FAILURE":
                raise RuntimeError(
                    "Execution resulted with status 'FAILURE'("
                    "no extra details about the errors are "
                    "available)."
                )

        LOGGER.info("%s finished with status: %s", str(action), action.status)

        return action

    @staticmethod
    @tenacity.retry(
        stop=tenacity.stop_after_attempt(5),
        wait=tenacity.wait_random(min=0.1, max=1),
        retry=tenacity.retry_if_exception_type(
            (ConnectionError, ConnectTimeout, RequestsConnectionError, DocumentError)
        ),
    )
    def refresh_action(
        current_action: resourceobject.ResourceObject,
    ) -> resourceobject.ResourceObject:
        """ Refreshes the action for polling purposes.

        Args:
            current_action: the current action

        Returns:
            the refreshed action

        """
        return current_action.refresh()


# pylint:disable=missing-class-docstring
class ColoredNoiseSimulationService(BaseService):
    _label = "colored_noise_simulations"
    _action_type = "colored_noise_simulation"

    def run(
        self, system: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Run the colored noise simulation.

        Args:
            system: Required.

        Returns:
            System: A system.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=system, name="Colored noise simulation {}".format(system.id)
        )


class CoherentSimulationService(BaseService):

    _label = "coherent_simulations"
    _action_type = "coherent_simulation"

    def run(
        self, system: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Run the coherent simulation.

        Args:
            system: Required.

        Returns:
            System: A system.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=system, name="Coherent simulation {}".format(system.id)
        )


class FilterFunctionService(BaseService):

    _label = "filter_functions"
    _action_type = "filter_function_evaluation"

    def calculate(
        self, filter_function: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Generate a filter function.

        Args:
            filter_function: Required.

        Returns:
            FilterFunction: A filter function.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=filter_function,
            name="Filter function calculation {}".format(filter_function.id),
        )


class T1DecayAssessmentService(BaseService):

    _label = "relaxation_infidelity"
    _action_type = "t1_decay_assessment"

    def calculate(
        self, system: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Measure the relaxation infidelities within a system.

        Args:
            system: Required.

        Returns:
            System: A system that contains relaxation infidelities.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=system, name="{}: {}".format(self.__class__.__name__, system.name)
        )


class NoiseFreeMetricsService(BaseService):

    _label = "noise_free_metrics"
    _action_type = "noise_free_metric_calculation"

    def calculate(
        self, target: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Calculate the infidelity and infidelity precision.

        Args:
            target: Required.

        Returns:
            Target: A target containing noise free metrics.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=target, name="Noise free metrics calculation {}".format(target.id)
        )


class RobustOptimizedSystemService(BaseService):

    _label = "robust_optimization"
    _action_type = "robust_optimize"

    def run(
        self, system: resourceobject.ResourceObject, num_optimizations: int = 20
    ) -> resourceobject.ResourceObject:
        """
        Args:
            system: Required.
            num_optimizations: Optional.

        Returns:
            System: An optimized system.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=system,
            name="{}: {}".format(self.__class__.__name__, system.name),
            num_optimizations=num_optimizations,
        )


class OptimalOptimizedSystemService(BaseService):

    _label = "optimal_optimization"
    _action_type = "optimal_optimize"

    def run(
        self, system: resourceobject.ResourceObject, num_optimizations: int = 20
    ) -> resourceobject.ResourceObject:
        """
        Args:
            system: Required.
            num_optimizations: Optional.

        Returns:
            System: An optimized system.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=system,
            name="{}: {}".format(self.__class__.__name__, system.name),
            num_optimizations=num_optimizations,
        )


class PiecewisePowerLawNoiseSpectralDensityService(BaseService):

    _label = "piecewise_power_law_noise_spectral_density"
    _action_type = "piecewise_power_law_noise_spectral_density_evaluation"

    def calculate(
        self, noise_spectral_density: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Plot interpolated values

        Args:
            noise_spectral_density: Required.

        Returns:
            NoiseSpectralDensity: A piecewise noise spectral density.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=noise_spectral_density,
            name="Noise spectral density calculation {}".format(
                noise_spectral_density.id
            ),
        )


class QuasiStaticFunctionService(BaseService):

    _label = "quasi_static_functions"
    _action_type = "quasi_static_function_evaluation"

    def calculate(
        self, quasi_static_function: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Calculate sampled_points

        Args:
            quasi_static_function: Required.

        Returns:
            QuasiStaticFunction: A quasi static function.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=quasi_static_function,
            name="Quasi static function calculation {}".format(
                quasi_static_function.id
            ),
        )


class NoiseCharacterizationService(BaseService):

    _label = "noise_characterization"
    _action_type = "noise_characterization_calculation"

    def calculate(
        self, noise_characterization: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Calculate Noise Characterization segments.

        Args:
            noise_characterization: Required.

        Returns:
            NoiseCharacterization: A noise characterization.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=noise_characterization,
            name="Noise characterization calculation {}".format(
                noise_characterization.id
            ),
        )


class NoiseReconstructionService(BaseService):

    _label = "noise_reconstructions"
    _action_type = "noise_reconstruction_calculation"

    def calculate(
        self, noise_reconstruction: resourceobject.ResourceObject
    ) -> resourceobject.ResourceObject:
        """Calculate a noise reconstruction.

        Args:
          noise_reconstruction: Required.

        Returns:
            NoiseReconstruction: A noise reconstruction.

        Raises:
            QctrlApiException
        """
        return self._run(
            resource=noise_reconstruction,
            name="Noise Reconstruction calculation {}".format(noise_reconstruction.id),
        )


class FlexibleOptimizedSystemService(BaseService):

    _label = "flexible_optimize"
    _action_type = "flexible_optimize"

    def run(
        self, graph: resourceobject.ResourceObject, num_optimizations: int = 20
    ) -> resourceobject.ResourceObject:
        """
        Args:
            graph: Required.
            num_optimizations: Optional.

        Returns:
            System: An optimized system.

        Raises:
            QctrlApiException
        """
        updated_graph = self._run(
            resource=graph,
            name="{}: {}".format(self.__class__.__name__, graph.id),
            num_optimizations=num_optimizations,
        )
        # FIXME bug with refreshing encoded fields; use refresh when solved pylint: disable=fixme
        updated_graph = self.api.graphs.find_by_id(updated_graph.id)
        return updated_graph


__all__ = [
    "ColoredNoiseSimulationService",
    "CoherentSimulationService",
    "FilterFunctionService",
    "T1DecayAssessmentService",
    "NoiseFreeMetricsService",
    "RobustOptimizedSystemService",
    "OptimalOptimizedSystemService",
    "PiecewisePowerLawNoiseSpectralDensityService",
    "QuasiStaticFunctionService",
    "NoiseCharacterizationService",
    "NoiseReconstructionService",
    "FlexibleOptimizedSystemService",
]
